@@ -1,67 +1,45 @@
-# -*- coding: utf-8 -*-
 """
 State tracking functionality for django models
 """
+
+from __future__ import annotations
+
 import inspect
-import sys
+from functools import partialmethod
 from functools import wraps
 
+from django.apps import apps as django_apps
 from django.db import models
+from django.db.models import Field
+from django.db.models.query_utils import DeferredAttribute
 from django.db.models.signals import class_prepared
-from django.utils.functional import curry
-from django_fsm.signals import pre_transition, post_transition
 
-try:
-    from django.apps import apps as django_apps
+from django_fsm.signals import post_transition
+from django_fsm.signals import pre_transition
 
-    def get_model(app_label, model_name):
-        app = django_apps.get_app_config(app_label)
-        return app.get_model(model_name)
-except ImportError:
-    from django.db.models.loading import get_model
-
-
-__all__ = ['TransitionNotAllowed', 'ConcurrentTransition',
-           'FSMFieldMixin', 'FSMField', 'FSMIntegerField',
-           'FSMKeyField', 'ConcurrentTransitionMixin', 'transition',
-           'can_proceed', 'has_transition_perm']
-
-if sys.version_info[:2] == (2, 6):
-    # Backport of Python 2.7 inspect.getmembers,
-    # since Python 2.6 ships buggy implementation
-    def __getmembers(object, predicate=None):
-        """Return all members of an object as (name, value) pairs sorted by name.
-        Optionally, only return members that satisfy a given predicate."""
-        results = []
-        for key in dir(object):
-            try:
-                value = getattr(object, key)
-            except AttributeError:
-                continue
-            if not predicate or predicate(value):
-                results.append((key, value))
-        results.sort()
-        return results
-    inspect.getmembers = __getmembers
-
-# South support; see http://south.aeracode.org/docs/tutorial/part4.html#simple-inheritance
-try:
-    from south.modelsinspector import add_introspection_rules
-except ImportError:
-    pass
-else:
-    add_introspection_rules([], [r"^django_fsm\.FSMField"])
-    add_introspection_rules([], [r"^django_fsm\.FSMIntegerField"])
-    add_introspection_rules([], [r"^django_fsm\.FSMKeyField"])
+__all__ = [
+    "TransitionNotAllowed",
+    "ConcurrentTransition",
+    "FSMFieldMixin",
+    "FSMField",
+    "FSMIntegerField",
+    "FSMKeyField",
+    "ConcurrentTransitionMixin",
+    "transition",
+    "can_proceed",
+    "has_transition_perm",
+    "GET_STATE",
+    "RETURN_VALUE",
+]
 
 
 class TransitionNotAllowed(Exception):
     """Raised when a transition is not allowed"""
 
     def __init__(self, *args, **kwargs):
-        self.object = kwargs.pop('object', None)
-        self.method = kwargs.pop('method', None)
-        super(TransitionNotAllowed, self).__init__(*args, **kwargs)
+        self.object = kwargs.pop("object", None)
+        self.method = kwargs.pop("method", None)
+        super().__init__(*args, **kwargs)
 
 
 class InvalidResultState(Exception):
@@ -76,7 +54,7 @@ class ConcurrentTransition(Exception):
     """
 
 
-class Transition(object):
+class Transition:
     def __init__(self, method, source, target, on_error, conditions, permission, custom):
         self.method = method
         self.source = source
@@ -93,14 +71,24 @@ class Transition(object):
     def has_perm(self, instance, user):
         if not self.permission:
             return True
-        elif callable(self.permission):
+        if callable(self.permission):
             return bool(self.permission(instance, user))
-        elif user.has_perm(self.permission, instance):
+        if user.has_perm(self.permission, instance):
             return True
-        elif user.has_perm(self.permission):
+        if user.has_perm(self.permission):
             return True
-        else:
-            return False
+        return False
+
+    def __hash__(self):
+        return hash(self.name)
+
+    def __eq__(self, other):
+        if isinstance(other, str):
+            return other == self.name
+        if isinstance(other, Transition):
+            return other.name == self.name
+
+        return False
 
 
 def get_available_FIELD_transitions(instance, field):
@@ -111,7 +99,7 @@ def get_available_FIELD_transitions(instance, field):
     curr_state = field.get_state(instance)
     transitions = field.transitions[instance.__class__]
 
-    for name, transition in transitions.items():
+    for transition in transitions.values():
         meta = transition._django_fsm
         if meta.has_transition(curr_state) and meta.conditions_met(instance, curr_state):
             yield meta.get_transition(curr_state)
@@ -134,10 +122,11 @@ def get_available_user_FIELD_transitions(instance, user, field):
             yield transition
 
 
-class FSMMeta(object):
+class FSMMeta:
     """
     Models methods transitions meta information
     """
+
     def __init__(self, field, method):
         self.field = field
         self.transitions = {}  # source -> Transition
@@ -145,14 +134,14 @@ class FSMMeta(object):
     def get_transition(self, source):
         transition = self.transitions.get(source, None)
         if transition is None:
-            transition = self.transitions.get('*', None)
+            transition = self.transitions.get("*", None)
         if transition is None:
-            transition = self.transitions.get('+', None)
+            transition = self.transitions.get("+", None)
         return transition
 
     def add_transition(self, method, source, target, on_error=None, conditions=[], permission=None, custom={}):
         if source in self.transitions:
-            raise AssertionError('Duplicate transition for {0} state'.format(source))
+            raise AssertionError(f"Duplicate transition for {source} state")
 
         self.transitions[source] = Transition(
             method=method,
@@ -161,7 +150,8 @@ class FSMMeta(object):
             on_error=on_error,
             conditions=conditions,
             permission=permission,
-            custom=custom)
+            custom=custom,
+        )
 
     def has_transition(self, state):
         """
@@ -170,10 +160,10 @@ class FSMMeta(object):
         if state in self.transitions:
             return True
 
-        if '*' in self.transitions:
+        if "*" in self.transitions:
             return True
 
-        if '+' in self.transitions and self.transitions['+'].target != state:
+        if "+" in self.transitions and self.transitions["+"].target != state:
             return True
 
         return False
@@ -186,24 +176,25 @@ class FSMMeta(object):
 
         if transition is None:
             return False
-        elif transition.conditions is None:
+
+        if transition.conditions is None:
             return True
-        else:
-            return all(map(lambda condition: condition(instance), transition.conditions))
+
+        return all(condition(instance) for condition in transition.conditions)
 
     def has_transition_perm(self, instance, state, user):
         transition = self.get_transition(state)
 
         if not transition:
             return False
-        else:
-            return transition.has_perm(instance, user)
+
+        return transition.has_perm(instance, user)
 
     def next_state(self, current_state):
         transition = self.get_transition(current_state)
 
         if transition is None:
-            raise TransitionNotAllowed('No transition from {0}'.format(current_state))
+            raise TransitionNotAllowed(f"No transition from {current_state}")
 
         return transition.target
 
@@ -211,59 +202,61 @@ class FSMMeta(object):
         transition = self.get_transition(current_state)
 
         if transition is None:
-            raise TransitionNotAllowed('No transition from {0}'.format(current_state))
+            raise TransitionNotAllowed(f"No transition from {current_state}")
 
         return transition.on_error
 
 
-class FSMFieldDescriptor(object):
+class FSMFieldDescriptor:
     def __init__(self, field):
         self.field = field
 
     def __get__(self, instance, type=None):
         if instance is None:
-            raise AttributeError('Can only be accessed via an instance.')
+            return self
         return self.field.get_state(instance)
 
     def __set__(self, instance, value):
         if self.field.protected and self.field.name in instance.__dict__:
-            raise AttributeError('Direct {0} modification is not allowed'.format(self.field.name))
+            raise AttributeError(f"Direct {self.field.name} modification is not allowed")
 
         # Update state
         self.field.set_proxy(instance, value)
         self.field.set_state(instance, value)
 
 
-class FSMFieldMixin(object):
+class FSMFieldMixin:
     descriptor_class = FSMFieldDescriptor
 
     def __init__(self, *args, **kwargs):
-        self.protected = kwargs.pop('protected', False)
+        self.protected = kwargs.pop("protected", False)
         self.transitions = {}  # cls -> (transitions name -> method)
         self.state_proxy = {}  # state -> ProxyClsRef
 
-        state_choices = kwargs.pop('state_choices', None)
-        choices = kwargs.get('choices', None)
+        state_choices = kwargs.pop("state_choices", None)
+        choices = kwargs.get("choices", None)
         if state_choices is not None and choices is not None:
-            raise ValueError('Use one of choices or state_choces value')
+            raise ValueError("Use one of choices or state_choices value")
 
         if state_choices is not None:
             choices = []
             for state, title, proxy_cls_ref in state_choices:
                 choices.append((state, title))
                 self.state_proxy[state] = proxy_cls_ref
-            kwargs['choices'] = choices
+            kwargs["choices"] = choices
 
-        super(FSMFieldMixin, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def deconstruct(self):
-        name, path, args, kwargs = super(FSMFieldMixin, self).deconstruct()
+        name, path, args, kwargs = super().deconstruct()
         if self.protected:
-            kwargs['protected'] = self.protected
+            kwargs["protected"] = self.protected
         return name, path, args, kwargs
 
     def get_state(self, instance):
-        return instance.__dict__[self.name]
+        # The state field may be deferred. We delegate the logic of figuring this out
+        # and loading the deferred field on-demand to Django's built-in DeferredAttribute class.
+        return DeferredAttribute(self).__get__(instance)
 
     def set_state(self, instance, state):
         instance.__dict__[self.name] = state
@@ -282,9 +275,10 @@ class FSMFieldMixin(object):
                 app_label = instance._meta.app_label
                 model_name = state_proxy
 
-            model = get_model(app_label, model_name)
+            model = django_apps.get_app_config(app_label).get_model(model_name)
+
             if model is None:
-                raise ValueError('No model found {0}'.format(state_proxy))
+                raise ValueError(f"No model found {state_proxy}")
 
             instance.__class__ = model
 
@@ -295,21 +289,26 @@ class FSMFieldMixin(object):
 
         if not meta.has_transition(current_state):
             raise TransitionNotAllowed(
-                "Can't switch from state '{0}' using method '{1}'".format(current_state, method_name),
-                object=instance, method=method)
+                f"Can't switch from state '{current_state}' using method '{method_name}'",
+                object=instance,
+                method=method,
+            )
         if not meta.conditions_met(instance, current_state):
             raise TransitionNotAllowed(
-                "Transition conditions have not been met for method '{0}'".format(method_name),
-                object=instance, method=method)
+                f"Transition conditions have not been met for method '{method_name}'", object=instance, method=method
+            )
 
         next_state = meta.next_state(current_state)
 
         signal_kwargs = {
-            'sender': instance.__class__,
-            'instance': instance,
-            'name': method_name,
-            'source': current_state,
-            'target': next_state
+            "sender": instance.__class__,
+            "instance": instance,
+            "name": method_name,
+            "field": meta.field,
+            "source": current_state,
+            "target": next_state,
+            "method_args": args,
+            "method_kwargs": kwargs,
         }
 
         pre_transition.send(**signal_kwargs)
@@ -317,10 +316,9 @@ class FSMFieldMixin(object):
         try:
             result = method(instance, *args, **kwargs)
             if next_state is not None:
-                if hasattr(next_state, 'get_state'):
-                    next_state = next_state.get_state(
-                        instance, transition, result,
-                        args=args, kwargs=kwargs)
+                if hasattr(next_state, "get_state"):
+                    next_state = next_state.get_state(instance, transition, result, args=args, kwargs=kwargs)
+                    signal_kwargs["target"] = next_state
                 self.set_proxy(instance, next_state)
                 self.set_state(instance, next_state)
         except Exception as exc:
@@ -328,8 +326,8 @@ class FSMFieldMixin(object):
             if exception_state:
                 self.set_proxy(instance, exception_state)
                 self.set_state(instance, exception_state)
-                signal_kwargs['target'] = exception_state
-                signal_kwargs['exception'] = exc
+                signal_kwargs["target"] = exception_state
+                signal_kwargs["exception"] = exc
                 post_transition.send(**signal_kwargs)
             raise
         else:
@@ -343,36 +341,46 @@ class FSMFieldMixin(object):
         """
         transitions = self.transitions[instance_cls]
 
-        for name, transition in transitions.items():
+        for transition in transitions.values():
             meta = transition._django_fsm
 
             for transition in meta.transitions.values():
                 yield transition
 
-    def contribute_to_class(self, cls, name, virtual_only=False):
+    def contribute_to_class(self, cls, name, **kwargs):
         self.base_cls = cls
 
-        super(FSMFieldMixin, self).contribute_to_class(cls, name, virtual_only=virtual_only)
+        super().contribute_to_class(cls, name, **kwargs)
         setattr(cls, self.name, self.descriptor_class(self))
-        setattr(cls, 'get_all_{0}_transitions'.format(self.name),
-                curry(get_all_FIELD_transitions, field=self))
-        setattr(cls, 'get_available_{0}_transitions'.format(self.name),
-                curry(get_available_FIELD_transitions, field=self))
-        setattr(cls, 'get_available_user_{0}_transitions'.format(self.name),
-                curry(get_available_user_FIELD_transitions, field=self))
+        setattr(cls, f"get_all_{self.name}_transitions", partialmethod(get_all_FIELD_transitions, field=self))
+        setattr(cls, f"get_available_{self.name}_transitions", partialmethod(get_available_FIELD_transitions, field=self))
+        setattr(
+            cls,
+            f"get_available_user_{self.name}_transitions",
+            partialmethod(get_available_user_FIELD_transitions, field=self),
+        )
 
         class_prepared.connect(self._collect_transitions)
 
     def _collect_transitions(self, *args, **kwargs):
-        sender = kwargs['sender']
+        sender = kwargs["sender"]
 
         if not issubclass(sender, self.base_cls):
             return
 
         def is_field_transition_method(attr):
-            return (inspect.ismethod(attr) or inspect.isfunction(attr)) \
-                and hasattr(attr, '_django_fsm') \
-                and attr._django_fsm.field in [self, self.name]
+            return (
+                (inspect.ismethod(attr) or inspect.isfunction(attr))
+                and hasattr(attr, "_django_fsm")
+                and (
+                    attr._django_fsm.field in [self, self.name]
+                    or (
+                        isinstance(attr._django_fsm.field, Field)
+                        and attr._django_fsm.field.name == self.name
+                        and attr._django_fsm.field.creation_counter == self.creation_counter
+                    )
+                )
+            )
 
         sender_transitions = {}
         transitions = inspect.getmembers(sender, predicate=is_field_transition_method)
@@ -387,15 +395,17 @@ class FSMField(FSMFieldMixin, models.CharField):
     """
     State Machine support for Django model as CharField
     """
+
     def __init__(self, *args, **kwargs):
-        kwargs.setdefault('max_length', 50)
-        super(FSMField, self).__init__(*args, **kwargs)
+        kwargs.setdefault("max_length", 50)
+        super().__init__(*args, **kwargs)
 
 
 class FSMIntegerField(FSMFieldMixin, models.IntegerField):
     """
     Same as FSMField, but stores the state value in an IntegerField.
     """
+
     pass
 
 
@@ -403,6 +413,7 @@ class FSMKeyField(FSMFieldMixin, models.ForeignKey):
     """
     State Machine support for Django model
     """
+
     def get_state(self, instance):
         return instance.__dict__[self.attname]
 
@@ -410,7 +421,34 @@ class FSMKeyField(FSMFieldMixin, models.ForeignKey):
         instance.__dict__[self.attname] = self.to_python(state)
 
 
-class ConcurrentTransitionMixin(object):
+class FSMModelMixin:
+    """
+    Mixin that allows refresh_from_db for models with fsm protected fields
+    """
+
+    def _get_protected_fsm_fields(self):
+        def is_fsm_and_protected(f):
+            return isinstance(f, FSMFieldMixin) and f.protected
+
+        protected_fields = filter(is_fsm_and_protected, self._meta.concrete_fields)
+        return {f.attname for f in protected_fields}
+
+    def refresh_from_db(self, *args, **kwargs):
+        fields = kwargs.pop("fields", None)
+
+        # Use provided fields, if not set then reload all non-deferred fields.0
+        if not fields:
+            deferred_fields = self.get_deferred_fields()
+            protected_fields = self._get_protected_fsm_fields()
+            skipped_fields = deferred_fields.union(protected_fields)
+
+            fields = [f.attname for f in self._meta.concrete_fields if f.attname not in skipped_fields]
+
+        kwargs["fields"] = fields
+        super().refresh_from_db(*args, **kwargs)
+
+
+class ConcurrentTransitionMixin:
     """
     Protects a Model from undesirable effects caused by concurrently executed transitions,
     e.g. running the same transition multiple times at the same time, or running different
@@ -435,16 +473,14 @@ class ConcurrentTransitionMixin(object):
     a rollback of all the changes that have been executed in an inconsistent (out of sync)
     state, thus practically negating their effect.
     """
+
     def __init__(self, *args, **kwargs):
-        super(ConcurrentTransitionMixin, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self._update_initial_state()
 
     @property
     def state_fields(self):
-        return filter(
-            lambda field: isinstance(field, FSMFieldMixin),
-            self._meta.fields
-        )
+        return filter(lambda field: isinstance(field, FSMFieldMixin), self._meta.fields)
 
     def _do_update(self, base_qs, using, pk_val, values, update_fields, forced_update):
         # _do_update is called once for each model class in the inheritance hierarchy.
@@ -454,15 +490,15 @@ class ConcurrentTransitionMixin(object):
         filter_on = filter(lambda field: field.model == base_qs.model, self.state_fields)
 
         # state filter will be used to narrow down the standard filter checking only PK
-        state_filter = dict((field.attname, self.__initial_states[field.attname]) for field in filter_on)
+        state_filter = {field.attname: self.__initial_states[field.attname] for field in filter_on}
 
-        updated = super(ConcurrentTransitionMixin, self)._do_update(
+        updated = super()._do_update(
             base_qs=base_qs.filter(**state_filter),
             using=using,
             pk_val=pk_val,
             values=values,
             update_fields=update_fields,
-            forced_update=forced_update
+            forced_update=forced_update,
         )
 
         # It may happen that nothing was updated in the original _do_update method not because of unmatching state,
@@ -471,36 +507,39 @@ class ConcurrentTransitionMixin(object):
         # INSERT if UPDATE fails.
         # Thus, we need to make sure we only catch the case when the object *is* in the DB, but with changed state; and
         # mimic standard _do_update behavior otherwise. Django will pick it up and execute _do_insert.
-        if not updated and base_qs.filter(pk=pk_val).exists():
+        if not updated and base_qs.filter(pk=pk_val).using(using).exists():
             raise ConcurrentTransition("Cannot save object! The state has been changed since fetched from the database!")
 
         return updated
 
     def _update_initial_state(self):
-        self.__initial_states = dict(
-            (field.attname, field.value_from_object(self)) for field in self.state_fields
-        )
+        self.__initial_states = {field.attname: field.value_from_object(self) for field in self.state_fields}
+
+    def refresh_from_db(self, *args, **kwargs):
+        super().refresh_from_db(*args, **kwargs)
+        self._update_initial_state()
 
     def save(self, *args, **kwargs):
-        super(ConcurrentTransitionMixin, self).save(*args, **kwargs)
+        super().save(*args, **kwargs)
         self._update_initial_state()
 
 
-def transition(field, source='*', target=None, on_error=None, conditions=[], permission=None, custom={}):
+def transition(field, source="*", target=None, on_error=None, conditions=[], permission=None, custom={}):
     """
-    Method decorator for mark allowed transitions
+    Method decorator to mark allowed transitions.
 
     Set target to None if current state needs to be validated and
-    has not changed after the function call
+    has not changed after the function call.
     """
+
     def inner_transition(func):
-        wrapper_installed, fsm_meta = True, getattr(func, '_django_fsm', None)
+        wrapper_installed, fsm_meta = True, getattr(func, "_django_fsm", None)
         if not fsm_meta:
             wrapper_installed = False
             fsm_meta = FSMMeta(field=field, method=func)
-            setattr(func, '_django_fsm', fsm_meta)
+            setattr(func, "_django_fsm", fsm_meta)
 
-        if isinstance(source, (list, tuple)):
+        if isinstance(source, (list, tuple, set)):
             for state in source:
                 func._django_fsm.add_transition(func, state, target, on_error, conditions, permission, custom)
         else:
@@ -525,34 +564,35 @@ def can_proceed(bound_method, check_conditions=True):
     Set ``check_conditions`` argument to ``False`` to skip checking
     conditions.
     """
-    if not hasattr(bound_method, '_django_fsm'):
-        raise TypeError('%s method is not transition' % bound_method.im_func.__name__)
+    if not hasattr(bound_method, "_django_fsm"):
+        raise TypeError(f"{bound_method.__func__.__name__} method is not transition")
 
     meta = bound_method._django_fsm
-    im_self = getattr(bound_method, 'im_self', getattr(bound_method, '__self__'))
-    current_state = meta.field.get_state(im_self)
+    self = bound_method.__self__
+    current_state = meta.field.get_state(self)
 
-    return meta.has_transition(current_state) and (
-        not check_conditions or meta.conditions_met(im_self, current_state))
+    return meta.has_transition(current_state) and (not check_conditions or meta.conditions_met(self, current_state))
 
 
 def has_transition_perm(bound_method, user):
     """
     Returns True if model in state allows to call bound_method and user have rights on it
     """
-    if not hasattr(bound_method, '_django_fsm'):
-        raise TypeError('%s method is not transition' % bound_method.im_func.__name__)
+    if not hasattr(bound_method, "_django_fsm"):
+        raise TypeError(f"{bound_method.__func__.__name__} method is not transition")
 
     meta = bound_method._django_fsm
-    im_self = getattr(bound_method, 'im_self', getattr(bound_method, '__self__'))
-    current_state = meta.field.get_state(im_self)
+    self = bound_method.__self__
+    current_state = meta.field.get_state(self)
 
-    return (meta.has_transition(current_state) and
-            meta.conditions_met(im_self, current_state) and
-            meta.has_transition_perm(im_self, current_state, user))
+    return (
+        meta.has_transition(current_state)
+        and meta.conditions_met(self, current_state)
+        and meta.has_transition_perm(self, current_state, user)
+    )
 
 
-class State(object):
+class State:
     def get_state(self, model, transition, result, args=[], kwargs={}):
         raise NotImplementedError
 
@@ -564,9 +604,7 @@ class RETURN_VALUE(State):
     def get_state(self, model, transition, result, args=[], kwargs={}):
         if self.allowed_states is not None:
             if result not in self.allowed_states:
-                raise InvalidResultState(
-                    '{} is not in list of allowed states\n{}'.format(
-                        result, self.allowed_states))
+                raise InvalidResultState(f"{result} is not in list of allowed states\n{self.allowed_states}")
         return result
 
 
@@ -579,7 +617,5 @@ class GET_STATE(State):
         result_state = self.func(model, *args, **kwargs)
         if self.allowed_states is not None:
             if result_state not in self.allowed_states:
-                raise InvalidResultState(
-                    '{} is not in list of allowed states\n{}'.format(
-                        result, self.allowed_states))
+                raise InvalidResultState(f"{result_state} is not in list of allowed states\n{self.allowed_states}")
         return result_state
